/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      FindFiles
Version:   1.0
ID:        ID_FINDFILES
Category:  Tool/Command
Date:      November 2003
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2000-2003.  All rights reserved.
Keywords:  find, file, directory, search
Short:     Used to find files and directories for the user.

###DESCRIPTION###
<p>The FindFiles class can be used either as a command to find and print file
lists, or can be integrated into an interface as a Find Files application.
Support is provided for setting wild-marked filters and content searching.
All you need to provide is a reference to an object capable of receiving text
output, so that the user can read the results of the search.  This DML based
example uses a FindFiles object to search for text files in the 'documents:'
directory, which include the word 'Athene':</p>

<pre>
  &lt;findfiles output=&quot;[text]&quot; location=&quot;documents:&quot; filter=&quot;*.txt&quot;
    filecontent=&quot;Athene&quot;/&gt;
</pre>

<p>When searching files for content, it is recommended that a filter is used
so that only a limited number of files are opened for searching.  If filtering
is not used then a content search can take a considerable amount of time to
complete, so try to avoid it.</p>

<p>This class can also be used to scan for specific files and perform an action
on each file as it is found.  To do this, initialise the objects you want to
call inside the findfiles object.  It is important that you use the <i>File</i>
field dynamically so that you can read the most recently found file-path for
each discovery cycle.  Here is an example that uses this feature to delete all
object files (*.o) from a directory tree.</p>

<pre>
  &lt;findfiles location=&quot;athene:sdk/classes/&quot; filter=&quot;*.o&quot;&gt;
    &lt;script static src="commands:delete" &amp;src="\[owner.file\]"/&gt;
  &lt;/findfiles&gt;
</pre>

<p>To see a working Find Files application, refer to the
tools:filesystem/findfiles/ directory, which will show you how to link user
input boxes to a FindFiles object.</p>

###END###

*****************************************************************************/

#define DEBUG

#include <pandora/system/all.h>
#include <pandora/files/all.h>
#include <pandora/graphics/text.h>
#include <pandora/modules/filesystem.h>
#include <pandora/modules/strings.h>
#include <pandora/functions/set.c>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_FINDFILES 1.0

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_FINDFILES, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2000-2004.  All rights reserved.",
   "January 2004",
   "FindFiles"
};

static struct KernelBase *KernelBase;
static struct FileSystemBase *FileSystemBase;
static struct StringsBase *StringsBase;
static OBJECTPTR modFile = NULL, modStrings = NULL;
static OBJECTPTR FindFilesClass = NULL;
static FIELD FID_SrcFile, FID_DestFile;

/*****************************************************************************
** Class definition.
*/

struct FindFiles {
   OBJECT_HEADER       /* Class header */
   OBJECTID OutputID;
   LONG     Static;
   LONG     Operation;
   LONG     Response;
   LONG     Flags;

   PRIVATE_FIELDS
   UBYTE Location[300];   /* The path to be searched */
   UBYTE Filter[40];      /* File name filter to apply to the search */
   UBYTE Content[60];     /* Content to look for in the searched files */
   UBYTE prvFile[160];    /* Refers to the latest file that has been found (used for callback purposes) */
   UBYTE Dest[300];
   BYTE  Stop;
   LONG  TotalChildren;      /* Total number of children belonging to the FindFiles object */
   STRING ConfirmScript;
   STRING ErrorScript;
   UBYTE *DataBuffer;
   struct ChildEntry ChildList[10];
};

enum {
   FOP_NONE=0,
   FOP_COPY,
   FOP_MOVE,
   FOP_DELETE
};

enum {
   RSP_NONE=0,
   RSP_CANCEL,
   RSP_YES,
   RSP_YESALL,
   RSP_NO
};

static ERROR GET_ConfirmScript(struct FindFiles *, STRING *);
static ERROR GET_Dest(struct FindFiles *, STRING *);
static ERROR GET_ErrorScript(struct FindFiles *, STRING *);
static ERROR GET_File(struct FindFiles *, STRING *);
static ERROR GET_FileContent(struct FindFiles *, STRING *);
static ERROR GET_Filter(struct FindFiles *, STRING *);
static ERROR GET_Location(struct FindFiles *, STRING *);

static ERROR SET_ConfirmScript(struct FindFiles *, STRING);
static ERROR SET_Dest(struct FindFiles *, STRING);
static ERROR SET_ErrorScript(struct FindFiles *, STRING);
static ERROR SET_FileContent(struct FindFiles *, STRING);
static ERROR SET_Filter(struct FindFiles *, STRING);
static ERROR SET_Location(struct FindFiles *, STRING);

#define FF_FORCEDIR    0x00000001
#define FF_INCLUDEROOT 0x00000002
#define FF_XML         0x00000004
#define FF_NOACTIVATE  0X00000008

static struct FieldDef FindFlags[] = {
   { "FORCEDIR",    FF_FORCEDIR },
   { "INCLUDEROOT", FF_INCLUDEROOT },
   { "XML",         FF_XML },
   { "NOACTIVATE",  FF_NOACTIVATE },
   { NULL, NULL }
};

static struct FieldDef OperationLookup[] = {
   /* Please keep this array in numerical order when adding new entries */
   { "NONE",   NULL },
   { "COPY",   FOP_COPY },
   { "MOVE",   FOP_MOVE },
   { "DELETE", FOP_DELETE },
   { NULL, NULL }
};

static struct FieldDef ResponseLookup[] = {
   { "CANCEL", RSP_CANCEL },
   { "YES",    RSP_YES },
   { "YESALL", RSP_YESALL },
   { "NO",     RSP_NO },
   { "NOALL",  RSP_CANCEL },
   { NULL, NULL }
};

static struct FieldArray FindFilesFields[] = {
   { "Output",        0, FDF_OBJECTID|FDF_RW,  0, NULL, NULL },
   { "Static",        0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "Operation",     0, FDF_LONG|FDF_LOOKUP|FDF_RW, (LONG)&OperationLookup, NULL, NULL },
   { "Response",      0, FDF_LONG|FDF_LOOKUP|FDF_RW, (LONG)&ResponseLookup, NULL, NULL },
   { "Flags",         0, FDF_LONGFLAGS|FDF_RW, (LONG)&FindFlags, NULL, NULL },
   /*** Virtual fields ***/
   { "ConfirmScript", 0, FDF_STRING|FDF_RW, 0, GET_ConfirmScript, SET_ConfirmScript },
   { "Dest",          0, FDF_STRING|FDF_RW, 0, GET_Dest,          SET_Dest },
   { "ErrorScript",   0, FDF_STRING|FDF_RW, 0, GET_ErrorScript,   SET_ErrorScript },
   { "File",          0, FDF_STRING|FDF_R,  0, GET_File,          NULL },
   { "FileContent",   0, FDF_STRING|FDF_RW, 0, GET_FileContent,   SET_FileContent },
   { "Filter",        0, FDF_STRING|FDF_RW, 0, GET_Filter,        SET_Filter },
   { "Location",      0, FDF_STRING|FDF_RW, 0, GET_Location,      SET_Location },
   { "Src",           0, FDF_STRING|FDF_RW, 0, GET_Location,      SET_Location },
    END_FIELD
};

/****************************************************************************/

static ERROR FIND_Activate(struct FindFiles *, APTR);
static ERROR FIND_ClosingTag(struct FindFiles *, APTR);
static ERROR FIND_Deactivate(struct FindFiles *, APTR);
static ERROR FIND_Free(struct FindFiles *, APTR);
static ERROR FIND_Init(struct FindFiles *, APTR);

static struct ActionArray FindFilesActions[] = {
   { AC_Activate,      FIND_Activate },
   { AC_ClosingTag,    FIND_ClosingTag },
   { AC_Deactivate,    FIND_Deactivate },
   { AC_Free,          FIND_Free },
   { AC_Init,          FIND_Init },
   { NULL, NULL }
};

/****************************************************************************/

#define MT_AddLocation -1

struct mtAddLocation {
   STRING Location;
};

static struct FunctionField argsAddLocation[] = { { "Location", ARG_STRING }, { NULL, NULL } };

static ERROR FIND_AddLocation(struct FindFiles *, struct mtAddLocation *);

static struct MethodArray FindFilesMethods[] = {
   { MT_AddLocation, FIND_AddLocation, "AddLocation", argsAddLocation, sizeof(struct mtAddLocation) },
   { NULL, NULL, NULL, NULL }
};

/*****************************************************************************
** Function prototypes.
*/

static LONG ContentMatch(struct FindFiles *, STRING, STRING, UBYTE *);
static void FileMatched(struct FindFiles *, STRING, STRING, OBJECTPTR, struct DirInfo *);
static void OutputText(struct FindFiles *, STRING);
static void ScanLocation(struct FindFiles *, STRING, OBJECTPTR);
static void UserConfirmation(struct FindFiles *, STRING, STRING);
static void UserError(struct FindFiles *, STRING, ERROR);
static ERROR FileCopy(struct FindFiles *, STRING, STRING);
static ERROR FileDelete(struct FindFiles *, STRING);
static ERROR FileMove(struct FindFiles *, STRING, STRING);

/*****************************************************************************
** User Instructions.
*/

static BYTE FindFilesUsage[] = {
"USAGE: findfiles [src]=filename [content]=filecontent [filter]=namefilter\n\
\n\
Use the FindFiles command to search for files in your system.  At\n\
a minimum you should supply the directory location from which you wish\n\
to start the search.  Optionally you can also search on file content\n\
and filter on file names.  Here is an example that searches for PNG\n\
files:\n\
\n\
  findfiles src=athene:pictures/ filter=*.png\n\
\n\
If you do not specify the starting location, the search will start\n\
from the athene: location."
};

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   /*** Open required modules ***/

   if (CreateObject(ID_MODULE, NULL, &modStrings, NULL,
      FID_Name|TSTRING,   "strings",
      FID_Version|TFLOAT, MODVERSION_STRINGS,
      TAGEND) IS ERR_Okay) {
      if (GetField(modStrings, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   if (CreateObject(ID_MODULE, NULL, &modFile, NULL,
      FID_Name|TSTRING,   "filesystem",
      FID_Version|TFLOAT, MODVERSION_FILESYSTEM,
      TAGEND) IS ERR_Okay) {
      if (GetField(modFile, FID_ModBase, FT_POINTER, &FileSystemBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   ResolveFields("SrcFile",  &FID_SrcFile,
                 "DestFile", &FID_DestFile,
                 TAGEND);

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&FindFilesClass, NULL,
      FID_BaseClassID|TLONG, ID_FINDFILES,
      FID_Version|TFLOAT,    VER_FINDFILES,
      FID_Name|TSTRING,      "FindFiles",
      FID_Category|TLONG,    CCF_COMMAND,
      FID_Usage|TSTRING,     FindFilesUsage,
      FID_Actions|TPTR,      FindFilesActions,
      FID_Methods|TPTR,      FindFilesMethods,
      FID_Fields|TPTR,       FindFilesFields,
      FID_Size|TLONG,        sizeof(struct FindFiles),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (FindFilesClass) { acFree(FindFilesClass); FindFilesClass = NULL; }
   if (modFile)        { acFree(modFile); modFile = NULL; }
   if (modStrings)     { acFree(modStrings); modStrings = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** Internal: IsDirectory()
** Short:    Returns TRUE if the string references a directory location.
*/

static BYTE IsDirectory(STRING Path)
{
   WORD i;

   if (!Path) return(FALSE);

   for (i=0; Path[i]; i++);
   if ((Path[i-1] IS '/') OR (Path[i-1] IS '\\') OR (Path[i-1] IS ':')) return(TRUE);
   return(FALSE);
}

/*****************************************************************************

###ACTION###
Name:  Activate
Short: Executes the search process.

###DESCRIPTION###
<p>Activating a FindFiles object causes it to start searching for files.  At
a minimum you must have set up the <a href="#Output">Output</a> and
<a href="#Location">Location</a> fields before calling this action, or it will
fail.  The results of the search will be sent to the object referred to by the
Output field.</p>

<p>This action does not return until it has completed the search process.</p>

###RESULT###
ERR_Okay:            The search executed successfully.
ERR_FieldNotSet:     The Output or Location fields were not set.
ERR_NewObject:       A File object could not be created.
ERR_Init:            A File object could not be initialised.
ERR_AllocMemory:     A memory allocation failed.
ERR_ExclusiveDenied: Access to the output object was denied.

###END###

*****************************************************************************/

#define BUFFER_SIZE 4096

static ERROR FIND_Activate(struct FindFiles *Self, APTR Void)
{
   struct DirInfo list;
   OBJECTPTR output;
   STRING buffer;
   ERROR error;
   WORD len, i, k;

   if (!Self->Location) {
      if (SetField(Self, FID_Location, FT_STRING, "athene:") != ERR_Okay) {
         OutputText(Self, "Error: You have not given me a location.");
         return(ERR_FieldNotSet);
      }
   }

   /*** Get a list of all children if we have not done so already ***/

   if (Self->TotalChildren IS -1) {
      Self->TotalChildren = ARRAYSIZE(Self->ChildList);
      ListChildren(Self->Head.UniqueID, Self->ChildList, &Self->TotalChildren);
   }

   /*** Set the default user response as 'yes' if it has not been set already ***/

   if (!Self->Response) Self->Response = RSP_YES;
   else {
      /*** Do not proceed any further if the current response setting is 'cancel' ***/

      if (Self->Response IS RSP_CANCEL) return(ERR_Okay);
   }

   /* If we are using the delete operation, run the confirmation script just
   ** once before we start deleting files (running the confirmation script per
   ** selected file would otherwise be very annoying).
   */

   if ((Self->Operation IS FOP_DELETE) AND (Self->ConfirmScript)) {
      if (Self->Response != RSP_YESALL) {
         UserConfirmation(Self, Self->Location, Self->Location);
      }
      if ((Self->Response != RSP_YES) AND (Self->Response != RSP_YESALL)) {
         DPrintF("FindFiles:","User cancelled deletion of this file - response %d.", Self->Response);
         return(ERR_Okay);
      }
   }

   /*** Access the object that will receive our output and clear its existing content ***/

   output = NULL;
   if (Self->OutputID) {
      if (AccessObject(Self->OutputID, 3000, &output) != ERR_Okay) {
         return(ObjectError(Self, ERH_Activate, ERR_AccessObject));
      }
      Action(AC_Clear, output, NULL);
   }

   DPrintF("~Activate()","[FindFiles:%d] Filter: %s, Location: %s, Content: %s, Operation: %s", GetUniqueID(Self), Self->Filter, Self->Location, Self->Content, OperationLookup[Self->Operation]);

   if (AllocMemory(BUFFER_SIZE, MEM_STRING|MEM_NOCLEAR, (APTR *)&buffer, NULL) IS ERR_Okay) {
      if (AllocMemory(BUFFER_SIZE, MEM_STRING|MEM_NOCLEAR, (APTR *)&Self->DataBuffer, NULL) IS ERR_Okay) {

         Self->Stop = FALSE;

         error = ERR_Okay;

         for (len=0; Self->Location[len]; len++) buffer[len] = Self->Location[len];
         buffer[len] = 0;

         if (IsDirectory(Self->Location) IS TRUE) {

            /*** Start the scanning process ***/

            k = StrLength(Self->Dest);
            if (Self->Flags & FF_INCLUDEROOT) {
               if ((buffer[len-1] IS '/') OR (buffer[len-1] IS '\\')) {
                  i = len-1;
                  while ((i > 0) AND (buffer[i-1] != ':') AND (buffer[i-1] != '/') AND (buffer[i-1] != '\\')) i--;
                  StrCopy(buffer+i, Self->Dest+k, sizeof(Self->Dest)-k);
               }
            }

            #ifndef DEBUG
               DebugState(FALSE, -1);
            #endif

            ScanLocation(Self, buffer, output);

            #ifndef DEBUG
               DebugState(TRUE, -1);
            #endif

            Self->Dest[k] = 0;
         }
         else {
            /*** Single file analysis ***/

            while ((len > 0) AND (buffer[len-1] != '/') AND (buffer[len-1] != ':') AND (buffer[len-1] != '\\')) len--;

            if ((!Self->Filter[0]) OR (StrCompare(Self->Filter, buffer+len, 0, STR_WILDCARD) IS ERR_Okay)) {
               if (ContentMatch(Self, buffer, Self->Content, Self->DataBuffer) IS TRUE) {
                  UBYTE filename[256];
                  StrCopy(buffer+len, filename, sizeof(filename));
                  buffer[len] = 0;
                  ClearMemory(&list, sizeof(list));
                  FileMatched(Self, buffer, filename, output, &list);
               }
            }
         }

         FreeMemory(Self->DataBuffer);
         Self->DataBuffer = NULL;
      }
      else error = ObjectError(Self, ERH_Activate, ERR_AllocMemory);

      FreeMemory(buffer);
   }
   else error = ObjectError(Self, ERH_Activate, ERR_AllocMemory);

   if (output) ReleaseObject(output);

   StepBack();
   return(error);
}

/*****************************************************************************

###METHOD###
Name:  AddLocation
Short: Adds a new directory or file to the list of locations to scan.

###DESCRIPTION###
<p>The FindFiles class supports the scanning of multiple files and directories
if you need to scan more than one location.  To do this, call the AddLocation
method for each location that you need to add.  The path must be fully
qualified so that the difference between files and directories is clear.</p>

<p>Note that if you set the Location field of the FindFiles object, any
multiple locations that you have added with this method will be lost.</p>

###INPUT###
STRING/Location: Refers to the directory or file that you want to add.

###RESULT###
ERR_Okay: The location was added to the list.
ERR_Args: Invalid arguments were specified.

###END###

*****************************************************************************/

static ERROR FIND_AddLocation(struct FindFiles *Self, struct mtAddLocation *Args)
{
   if ((!Args) OR (!Args->Location)) return(ERR_Args);



   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  ClosingTag
Short: If the object is non-static, this action will perform the file search and then the object will self-destruct.
###END###

*****************************************************************************/

static ERROR FIND_ClosingTag(struct FindFiles *Self, APTR Void)
{
   Self->TotalChildren = ARRAYSIZE(Self->ChildList);
   ListChildren(Self->Head.UniqueID, Self->ChildList, &Self->TotalChildren);

   if (Self->Static IS FALSE) {
      Action(AC_Activate, Self, NULL);
      Action(AC_Free, Self, NULL);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Deactivate
Short: Immediaetly stops the find-file process if it is active.
###END###

*****************************************************************************/

static ERROR FIND_Deactivate(struct FindFiles *Self, APTR Void)
{
   Self->Stop = TRUE;
   return(ERR_Okay);
}

/*****************************************************************************
** FindFiles: Free
*/

static ERROR FIND_Free(struct FindFiles *Self, APTR Void)
{
   if (Self->ConfirmScript) {
      FreeMemory(Self->ConfirmScript);
      Self->ConfirmScript = NULL;
   }

   if (Self->ErrorScript) {
      FreeMemory(Self->ErrorScript);
      Self->ErrorScript = NULL;
   }

   return(ERR_Okay);
}

/*****************************************************************************
** FindFiles: Init
*/

static ERROR FIND_Init(struct FindFiles *Self, APTR Void)
{
   Self->TotalChildren = -1;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   ConfirmScript
Short:  Refers to a confirmation script to be used when performing special operations.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>When using the move, copy or delete operation with a FindFiles object,
you should set a confirmation script so that special instances can be handled,
such as overwriting a file.  The script in question will be expected to open
a window and present a dialog box to the user.  After opening the window, the
script waits for the user to respond before returning to the FindFiles object
(otherwise it will assume that the user has cancelled the operation).  You
can do this by using a Wait object.</p>

<p>The script can confirm the user response with the FindFiles object by
writing to the Response field.  Acceptable responses are YES, YESALL, NO
and CANCEL, depending on the operation in use.  Refer to the Response
field for more information.</p>

###SEE ALSO###
Field: Operation, Response
Class: Wait

###END###

*****************************************************************************/

static ERROR GET_ConfirmScript(struct FindFiles *Self, STRING *Value)
{
   if ((*Value = Self->ConfirmScript)) {
      *Value = Self->ConfirmScript;
      return(ERR_Okay);
   }
   else return(ERR_FieldNotSet);

   return(ERR_Okay);
}

static ERROR SET_ConfirmScript(struct FindFiles *Self, STRING Value)
{
   LONG i;

   if (Self->ConfirmScript) { FreeMemory(Self->ConfirmScript); Self->ConfirmScript = NULL; }

   if ((Value) AND (*Value)) {
      for (i=0; Value[i]; i++);
      if (AllocMemory(i+1, MEM_STRING|MEM_NOCLEAR, (void **)&Self->ConfirmScript, NULL) IS ERR_Okay) {
         for (i=0; Value[i]; i++) Self->ConfirmScript[i] = Value[i];
         Self->ConfirmScript[i] = 0;
         return(ERR_Okay);
      }
      else return(ObjectError(Self, ERH_SetField, ERR_AllocMemory));
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Dest
Short:  Specifies a directory destination; required when using special operations.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>When using certain special operations such as the copy or move modes with
a FindFiles object, you also need to set a destination path.  The path must
be fully qualified (ending with a trailing slash or colon as appropriate).
If the destination path does not exist, then it may be automatically created
by the system when you activate the find-file process.</p>

###END###

*****************************************************************************/

static ERROR GET_Dest(struct FindFiles *Self, STRING *Value)
{
   if (Self->Dest[0]) {
      *Value = Self->Dest;
      return(ERR_Okay);
   }
   else return(ERR_FieldNotSet);
}

static ERROR SET_Dest(struct FindFiles *Self, STRING Value)
{
   WORD i;

   if (Value) {
      for (i=0; (i < sizeof(Self->Dest)-1) AND (Value[i]); i++) Self->Dest[i] = Value[i];
      Self->Dest[i] = 0;
   }
   else Self->Dest[0] = 0;

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   ErrorScript
Short:  Refers to an error script that can be used for reporting errors to the user.
Type:   STRING
Status: Get/Set

###DESCRIPTION###

###END###

*****************************************************************************/

static ERROR GET_ErrorScript(struct FindFiles *Self, STRING *Value)
{
   if ((*Value = Self->ErrorScript)) {
      *Value = Self->ErrorScript;
      return(ERR_Okay);
   }
   else return(ERR_FieldNotSet);
}

static ERROR SET_ErrorScript(struct FindFiles *Self, STRING Value)
{
   LONG i;

   if (Self->ErrorScript) { FreeMemory(Self->ErrorScript); Self->ErrorScript = NULL; }

   if ((Value) AND (*Value)) {
      for (i=0; Value[i]; i++);
      if (AllocMemory(i+1, MEM_STRING|MEM_NOCLEAR, (void **)&Self->ErrorScript, NULL) IS ERR_Okay) {
         for (i=0; Value[i]; i++) Self->ErrorScript[i] = Value[i];
         Self->ErrorScript[i] = 0;
         return(ERR_Okay);
      }
      else return(ObjectError(Self, ERH_SetField, ERR_AllocMemory));
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   File
Short:  Refers to the most recently found file (used for callback purposes).
Status: Get
Type:   STRING

###DESCRIPTION###
<p>The File field can be read during the find-file process if the callback
feature is being used.  Note that this field can refer to directories as well
as files (directories can always be detected by checking for a slash or colon
character at the end of the string).</p>

###END###

*****************************************************************************/

static ERROR GET_File(struct FindFiles *Self, STRING *Value)
{
   if ((*Value = Self->prvFile)) return(ERR_Okay);
   else return(ERR_FieldNotSet);
}

/*****************************************************************************

###FIELD###
Name:   FileContent
Short:  If the contents of each file should be searched, apply the search string to this field.
Status: Read/Write
Type:   STRING

###DESCRIPTION###
<p>If the content of each file is to be searched, set this field to the
string that you want to use when searching the file content.  Files that
contain the string will be listed in the FindFiles output (on the condition
that they also passed the filter test).</p>

<p>Note that searching the content of each file increases the amount of time
required to do the search by a considerable level.  For this reason it is
important to use the content searching feature in conjunction with file
filtering.</p>

###END###

*****************************************************************************/

static ERROR GET_FileContent(struct FindFiles *Self, STRING *Value)
{
   if (Self->Content[0]) {
      *Value = Self->Content;
      return(ERR_Okay);
   }
   else {
      *Value = NULL;
      return(ERR_FieldNotSet);
   }
}

static ERROR SET_FileContent(struct FindFiles *Self, STRING Value)
{
   WORD i;

   if (Value) {
      for (i=0; (i < sizeof(Self->Content)-1) AND (Value[i]); i++) Self->Content[i] = Value[i];
      Self->Content[i] = 0;
   }
   else Self->Content[0] = 0;

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Filter
Short:  A file filter string can be applied by writing to this field.
Status: Read/Write
Type:   STRING

###DESCRIPTION###
<p>To define a file filter, set this field using standard wild-card values.
Both the asterisk and question-mark characters are accepted as wild-wards,
while the OR operator is reserved for use in a future update.  Here are some
filter examples:</p>

<pre>
   DML files:                             *.dml
   Files with extensions:                 *.*
   DML files with short names:            ???.dml
   Files where "b" is a second character: ?b*
   Files starting with a, ending in b:    a*b
</pre>

<p>File filters are not case sensitive.  If you do not use wild-cards then
the FindFiles object will scan for exact matches to the string you specify.</p>

###END###

*****************************************************************************/

static ERROR GET_Filter(struct FindFiles *Self, STRING *Value)
{
   if (Self->Filter[0]) {
      *Value = Self->Filter;
      return(ERR_Okay);
   }
   else {
      *Value = NULL;
      return(ERR_FieldNotSet);
   }
}

static ERROR SET_Filter(struct FindFiles *Self, STRING Value)
{
   WORD i;

   if (Value) {
      for (i=0; (i < sizeof(Self->Filter)-1) AND (Value[i]); i++) Self->Filter[i] = Value[i];
      Self->Filter[i] = 0;
   }
   else Self->Filter[0] = 0;

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Location
Short:  The directory location to be searched must be specified here.
Status: Write
Type:   STRING

###DESCRIPTION###
<p>The directory that is to be searched for files must be specified here.
Note that searching is recursive, so everything within the directory you
specify is also searched.  Searching from certain root directories (such as
"athene:" can result in long, processor intensive searching.  For this reason
it is always best to pin-point the location rather than to search from a
root directory if possible.</p>

###END###

*****************************************************************************/

static ERROR GET_Location(struct FindFiles *Self, STRING *Value)
{
   if (Self->Location[0]) {
      *Value = Self->Location;
      return(ERR_Okay);
   }
   else {
      *Value = NULL;
      return(ERR_FieldNotSet);
   }
   return(ERR_Okay);
}

static ERROR SET_Location(struct FindFiles *Self, STRING Value)
{
   WORD i;

   if ((Value) AND (*Value)) {
      for (i=0; (i < sizeof(Self->Location)-2) AND (Value[i]); i++) Self->Location[i] = Value[i];
      if (Self->Flags & FF_FORCEDIR) if ((Value[i-1] != '/') AND (Value[i-1] != ':') AND (Value[i-1] != '\\')) Self->Location[i++] = '/';
      Self->Location[i] = 0;
   }
   else Self->Location[0] = 0;

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Output
Short:  This field must be set to an object that can receive text output.
Status: Read/Write
Type:   OBJECTID

###DESCRIPTION###
<p>The Output field must be set so that a FindFiles object has a place to
send the results of a search to.  The Output object that you specify must
accept text-based information via data channels in order for the data to be
interpreted correctly.  If the Output object does not support text based
information, the FindFiles object will still perform the search but the user
will not receive any text based output.</p>

<p>It is recommended that this field points to an object that is a member
of the Text class.</p>

###SEE ALSO###
Class: Text

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Static
Short:  Set to TRUE to make the object static.
Status: Read/Write
Type:   OBJECTID

###DESCRIPTION###
<p>By default, a FindFiles object will execute itself and then self-destruct
when a closing tag is received.  If you would rather that the object stays in
the system, set this field to TRUE.  If you do this, the only way to get a
FindFiles object to perform is to call the Activate() action.</p>

###END###

*****************************************************************************/

/*****************************************************************************
** Function: OutputText()
*/

static void OutputText(struct FindFiles *Self, STRING Buffer)
{
   struct acDataChannel channel;

   if (Self->OutputID) {
      /*** Send the data to the output object ***/

      ActionMsg(AC_Clear, Self->OutputID, NULL);

      channel.ObjectID     = GetUniqueID(Self);
      channel.DataType     = DATA_TEXT;
      channel.Version      = 1;
      channel.TotalEntries = 1;
      channel.Buffer       = Buffer;
      channel.Size         = StrLength(Buffer) + 1;
      ActionMsg(AC_DataChannel, Self->OutputID, &channel);
   }
}

/*****************************************************************************
** Internal: ContentMatch()
*/

static LONG ContentMatch(struct FindFiles *Self, STRING Location, STRING Content, UBYTE *DataBuffer)
{
   struct acRead read;
   struct File *file;
   LONG size;
   WORD k, j;

   if ((!Content) OR (!*Content)) return(TRUE);

   if (CreateObject(ID_FILE, NULL, &file, NULL,
      FID_Location|TSTRING, Location,
      FID_Flags|TLONG,      FL_READ,
      TAGEND) IS ERR_Okay) {

      GetField(file, FID_Size, FT_LONG, &size);
      k = 0;
      while (size > 0) {
         read.Buffer = DataBuffer;
         if (size < BUFFER_SIZE) read.Length = size;
         else read.Length = BUFFER_SIZE;
         Action(AC_Read, file, &read);
         for (j=0; (j < size) AND (j < BUFFER_SIZE); j++) {
            if (lcase(DataBuffer[j]) != lcase(Content[k])) k = 0;
            else {
               k++;
               if (!Content[k]) {
                  acFree(file);
                  return(TRUE);
               }
            }
         }

         ProcessMessages(NULL, NULL);
         if (Self->Stop IS TRUE) break;

         size -= BUFFER_SIZE;
      }
      acFree(file);
   }

   return(FALSE);
}

/*****************************************************************************
** Internal: FileMatched()
*/

static void FileMatched(struct FindFiles *Self, STRING Path, STRING Filename, OBJECTPTR Output, struct DirInfo *Info)
{
   UBYTE buffer[256];
   OBJECTPTR object;
   WORD j, pathlen, filelen, fullpath_len;

   for (pathlen=0; Path[pathlen]; pathlen++);
   for (filelen=0; Filename[filelen]; filelen++);

   fullpath_len = StrCopy(Path, Self->prvFile, sizeof(Self->prvFile));
   StrCopy(Filename, Self->prvFile + fullpath_len, sizeof(Self->prvFile) - fullpath_len);

   /*** Print the file to the output object ***/

   if (Output) {
      if (Self->Flags & FF_XML) {
         StrFormat(buffer, sizeof(buffer), "<file icon=\"%s\">%s<folder>%s</folder><size sort=\"%.10d\">%d</size><date>%4d%2d%2d %d:%d:%d</date></file>", GetFileIcon(Self->prvFile), Filename, Path, (LONG)Info->Size, (LONG)Info->Size, Info->Time.Year, Info->Time.Month, Info->Time.Day, Info->Time.Hour, Info->Time.Minute, Info->Time.Second);
         ActionTags(AC_DataChannel, Output, Self->Head.UniqueID, DATA_XML, 1, buffer, StrLength(buffer)+1, 1);
      }
      else ActionTags(AC_DataChannel, Output, Self->Head.UniqueID, DATA_TEXT, 1, Self->prvFile, StrLength(Self->prvFile) + 1, 1);
   }

   if ((Self->TotalChildren > 0) AND (!(Self->Flags & FF_NOACTIVATE))) {
      /* Inform child objects about the file we've discovered.  Each object
      ** will be able to determine the currently active file by reading our
      ** File field.
      */

      for (j=0; j < Self->TotalChildren; j++) {
         if (AccessObject(Self->ChildList[j].ObjectID, 0, &object) IS ERR_Okay) {
            acActivate(object);
            ReleaseObject(object);
         }
      }
   }

   if (Self->Operation) {
      switch (Self->Operation) {
         case FOP_DELETE:
            FileDelete(Self, Self->prvFile);
            break;

         case FOP_COPY:
            FileCopy(Self, Self->prvFile, Filename);
            break;

         case FOP_MOVE:
            FileMove(Self, Self->prvFile, Filename);
            break;
      }
   }
}

/*****************************************************************************
** Internal: ScanLocation()
**
** Buffer - The file path currently being scanned.
*/

static void ScanLocation(struct FindFiles *Self, STRING Buffer, OBJECTPTR output)
{
   struct DirInfo *info;
   LONG k, bufferlen, count;
   WORD root, len, insert, dest_created;
   ERROR error;

   if (Self->Dest[0]) DPrintF("ScanLocation()","%s TO %s", Buffer, Self->Dest);
   else DPrintF("ScanLocation()","%s", Buffer);

   if ((Buffer[0] IS ':') OR (!Buffer[0])) root = TRUE;
   else root = FALSE;

   if (Self->Stop IS TRUE) return;

   for (len=0; Buffer[len]; len++);

   count = 0;
   dest_created = FALSE;
   if (OpenDirectory(Buffer, RDF_FILE|RDF_SIZE|RDF_DATE|RDF_PERMISSIONS, &info) IS ERR_Okay) {
      while (ScanDirectory(info) IS ERR_Okay) {
         count++;
         ProcessMessages(NULL, NULL);

         for (k=0; info->Name[k]; k++) Buffer[len+k] = info->Name[k];
         Buffer[len+k] = 0;
         bufferlen = len+k;

         /*** Ensure that the filename matches our filter ***/

         if (StrCompare(Self->Filter, info->Name, 0, STR_WILDCARD) != ERR_Okay) {
            #ifdef DEBUG
               DPrintF("ScanDir:","File \"%s\" does not match filter \"%s\"", info->Name, Self->Filter);
            #endif
            continue;
         }

         if (ContentMatch(Self, Buffer, Self->Content, Self->DataBuffer) IS FALSE) {
            #ifdef DEBUG
               DPrintF("ScanDir:","Content of file %s is not a match.", info->Name);
            #endif
            continue;
         }

         if (Self->Operation IS FOP_MOVE) {
            /* The destination directory must exist before moving a file */
            if (dest_created IS FALSE) {
               dest_created = TRUE;
               if (IsDirectory(Self->Dest)) MakeDirectory(Self->Dest, info->Permissions | PERMIT_USER);
            }
         }

         Buffer[len] = 0;

         FileMatched(Self, Buffer, info->Name, output, info);

         if (Self->Response IS RSP_CANCEL) Self->Stop = TRUE;

         if (Self->Stop IS TRUE) break;
      }

      Buffer[len] = 0;
      CloseDirectory(info);
   }

   if (Self->Stop IS TRUE) {
      #ifdef DEBUG
      DPrintF("ScanDirectory:","Stopping search.");
      #endif
      return;
   }

   /*** Recursively scan directory content ***/

   if (OpenDirectory(Buffer, RDF_DIRECTORY|RDF_QUALIFY, &info) IS ERR_Okay) {

      /* If there is a destination path given (for move/copy operations),
      ** we'll need to adjust it for each scanned sub-directory.  This sub-
      ** routine will calculate the insertion point for new directories.
      */

      insert = -1;
      if (Self->Dest[0]) {
         for (insert=0; Self->Dest[insert]; insert++);
         if ((Self->Dest[insert-1] != '/') AND (Self->Dest[insert-1] != '\\') AND
             (Self->Dest[insert-1] != ':')) {
            insert = -1;
         }
      }

      while (ScanDirectory(info) IS ERR_Okay) {
         count++;
         if (info->Flags & RDF_LINK) continue; /* Do not scan symbolicly linked directories (possible recursion) */

         ProcessMessages(NULL, NULL);

         if (root IS TRUE) {
            for (k=0; info->Name[k]; k++) Buffer[k] = info->Name[k];
            Buffer[k] = 0;
         }
         else {
            for (k=0; info->Name[k]; k++) Buffer[len+k] = info->Name[k];
            Buffer[len+k] = 0;
         }

         if (insert != -1) {
            StrCopy(info->Name, Self->Dest+insert, sizeof(Self->Dest) - insert);
         }

         ScanLocation(Self, Buffer, output);
         if (Self->Stop IS TRUE) break;

         /*** Put the destination path back to its original string ***/

         if (insert != -1) Self->Dest[insert] = 0;
      }
      CloseDirectory(info);
   }
   #ifdef DEBUG
   else DPrintF("ScanDirectory()","No folders in %s", Buffer);
   #endif

   Buffer[len] = 0;

   if ((!count) AND ((Self->Operation IS FOP_COPY) OR (Self->Operation IS FOP_MOVE))) {
      /*** The source directory is completely empty (irrespective of applied filters) ***/
      if (IsDirectory(Self->Dest)) {
         /* Use standard file system functions in order to preserve permissions.  The
         ** trailing slash is stripped from the end of the directory in order to prevent
         ** copy and move operations creating doubled-up instances of empty directories.
         */
         BYTE save;
         for (len=0; Self->Dest[len]; len++);
         save = Self->Dest[len-1];
         Self->Dest[len-1] = 0;

         //DPrintF("ScanDirectory:","Empty dir \"%s\" TO \"%s\"", Buffer, Self->Dest);

         if (Self->Operation IS FOP_COPY) error = CopyLocation(Buffer, Self->Dest);
         else error = MoveLocation(Buffer, Self->Dest);

         Self->Dest[len-1] = save;

         if (error != ERR_Okay) {
            DPrintF("ScanDirectory:","Error %d returned.", error);
            Self->Stop = TRUE;
            Self->Response = RSP_CANCEL;
            UserError(Self, Buffer, error);
            return;
         }
      }

      ProcessMessages(NULL, NULL); /* Message check required if the directory was empty */
   }
   else {
      /* For the move and delete operations, we delete the source
      ** directory when we are not using file filters.
      */

      if ((Self->Operation IS FOP_DELETE) AND (!Self->Filter[0]) AND (!Self->Content[0])) {
         DeleteFile(Buffer);
      }
      else if ((Self->Operation IS FOP_MOVE) AND (!Self->Filter[0]) AND (!Self->Content[0])) {
         DeleteFile(Buffer);
      }

      if (!count) ProcessMessages(NULL, NULL); /* Message check required if the directory was empty */
   }
}

/*****************************************************************************
** Internal: FileCopy()
**
** Location - The complete file path, including the file name.
** File - The name of the file only.
*/

static ERROR FileCopy(struct FindFiles *Self, STRING Location, STRING File)
{
   UBYTE dest[300];
   LONG type;
   ERROR error;
   WORD i;

   if (!Self->Dest[0]) return(ERR_Okay);

   /*** Determine the destination file ***/

   i = StrCopy(Self->Dest, dest, sizeof(dest)-1);

   if ((dest[i-1] IS '/') OR (dest[i-1] IS '\\') OR (dest[i-1] IS ':')) {
      i += StrCopy(File, dest+i, sizeof(dest)-i);
   }

   /* Unless there has been a yes-all response, run the confirmation script
   ** if the file exists at the destination location.
   */

   if ((Self->ConfirmScript) AND (Self->Response != RSP_YESALL)) {
      if ((AnalyseLocation(dest, &type) IS ERR_Okay) AND ((type IS LOC_FILE) OR (type IS LOC_DIRECTORY))) {
         UserConfirmation(Self, Location, dest);
      }
   }

   if ((Self->Response != RSP_YES) AND (Self->Response != RSP_YESALL)) {
      DPrintF("FindFiles:","Skipping file %s, response %d", Location, Self->Response);
      return(ERR_Okay);
   }

   /*** Copy the file ***/

   error = CopyLocation(Location, dest);

   /* If an error occurred during the copy operation, tell the user and
   ** cancel the rest of the copy process.  The possibility of a continuation
   ** (skipping of the error) is not allowed here.
   */

   if (error) {
      Self->Stop = TRUE;
      Self->Response = RSP_CANCEL;
      UserError(Self, Location, error);
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: FileDelete()
*/

static ERROR FileDelete(struct FindFiles *Self, STRING Location)
{
   ERROR error;

   if ((Self->Response IS RSP_CANCEL) OR (Self->Stop IS TRUE)) return(ERR_Okay);

   error = DeleteFile(Location);

   /* If an error occurs during the delete operation, inform the user but
   ** do not stop the process as the most likely problem is that it is just
   ** a permissions issue.
   */

   if (error) UserError(Self, Location, error);

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: FileMove()
*/

static ERROR FileMove(struct FindFiles *Self, STRING Location, STRING File)
{
   UBYTE dest[300];
   LONG type;
   ERROR error;
   WORD i;

   DPrintF("FileMove()","%s, File: %s", Location, File);

   if (!Self->Dest[0]) return(ERR_Okay);

   /*** Determine the destination file ***/

   i = StrCopy(Self->Dest, dest, sizeof(dest)-1);

   if ((dest[i-1] IS '/') OR (dest[i-1] IS '\\') OR (dest[i-1] IS ':')) {
      i += StrCopy(File, dest+i, sizeof(dest)-i);
   }

   /* Unless there has been a yes-all response, run the confirmation script
   ** if the file exists at the destination location.
   */

   if ((Self->ConfirmScript) AND (Self->Response != RSP_YESALL)) {
      if ((AnalyseLocation(dest, &type) IS ERR_Okay) AND ((type IS LOC_FILE) OR (type IS LOC_DIRECTORY))) {
         UserConfirmation(Self, Location, dest);
      }
   }

   if ((Self->Response != RSP_YES) AND (Self->Response != RSP_YESALL)) {
      DPrintF("FindFiles:","Skipping file %s, response %d", Location, Self->Response);
      return(ERR_Okay);
   }

   /*** Move the file ***/

   error = MoveLocation(Location, dest);

   /* If an error occurred during the copy operation, tell the user and
   ** cancel the rest of the copy process.  The possibility of a continuation
   ** (skipping of the error) is not allowed here.
   */

   if (error) {
      Self->Stop = TRUE;
      Self->Response = RSP_CANCEL;
      UserError(Self, Location, error);
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: UserConfirmation()
*/

static void UserConfirmation(struct FindFiles *Self, STRING SrcFile, STRING DestFile)
{
   OBJECTPTR script;

   if (Self->Response IS RSP_YESALL) return;

   if (!Self->ConfirmScript) {
      Self->Response = RSP_YES;
      return;
   }

   Self->Response = RSP_CANCEL;
   if (CreateObject(ID_SCRIPT, NF_CHILD, &script, NULL,
         FID_Location|TSTRING,   Self->ConfirmScript,
         FID_SrcFile|TUNLISTED,  SrcFile,
         FID_DestFile|TUNLISTED, DestFile,
         TAGEND) IS ERR_Okay) {
      acActivate(script);
      acFree(script);
   }
}

/*****************************************************************************
** Internal: UserError()
*/

static void UserError(struct FindFiles *Self, STRING Location, ERROR ErrorCode)
{
   OBJECTPTR script;

   if (!Self->ErrorScript) return;

   if (CreateObject(ID_SCRIPT, NF_CHILD, &script, NULL,
         FID_Location|TSTRING, Self->ErrorScript,
         FID_File|TUNLISTED,   Location,
         TAGEND) IS ERR_Okay) {
      acActivate(script);
      acFree(script);
   }
}
