/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Set
Version:   1.0
Status:    Stable
Date:      December 2001
ID:        ID_SET
Category:  Command
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2000-2001.  All rights reserved.
Keywords:  set, field, save, write, object

###DESCRIPTION###
<p>The Set class provides a way to update the field values of any public
object during script execution.  It can also be made static, which allows
the Set object to update field values during run-time, whenever it is
activated.  To create a valid Set object, tell it what object to interact
with and then list the fields that you want set using the unlisted field
modifier ('&amp;' character).  If you want the object to be static, make
sure that you use the static keyword before listing the fields, or the
Set object will pass on the new field values immediately.</p>

<p>The following  example illustrates how to update the x and y values of
a render object:</p>

<pre>
   &lt;render name="area" x="100" y="200" width="50" height="60"/&gt;
   &lt;set object="[area]" &amp;x="=[area.x]+50" &amp;y="100"/&gt;
</pre>

<p>If necessary, Set objects that have been defined as static can also
perform run-time object translations and calculations.  To activate
this feature, simply enclose the field setting in '$' characters, e.g:</p>

<pre>
   &lt;set static object="[area]" &amp;x="$=[area.x]+50$" &amp;y="100"/&gt;
</pre>

###END###

*****************************************************************************/

#include <pandora/system/all.h>
#include <pandora/modules/strings.h>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_SET 1.0

EXPORT struct ModHeader ModHeader = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_SET, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2000-2001.  All rights reserved.",
   "April 2001",
   "Set"
};

static struct KernelBase  *KernelBase = 0;
static struct StringsBase *StringsBase = 0;
static OBJECTPTR modStrings = 0;
static OBJECTPTR SetClass = 0;

/*****************************************************************************
** Class definition.
*/

struct Set {
   OBJECT_HEADER       /* Class header */
   LONG Static;

   PRIVATE_FIELDS
   OBJECTID ObjectID;
   BYTE ObjectReference[100];
   STRING Fields[20];
};

static ERROR GET_Object(struct Set *, struct Variable *);
static ERROR SET_Object(struct Set *, struct Variable *);

static struct FieldArray _SetFields[] = {
   { "Static", 0, FDF_LONG|FDF_INIT,   0, NULL, NULL },
   /*** Virtual fields ***/
   { "Object", 0, FDF_VARIABLE|FDF_STRING|FDF_RW, 0, GET_Object, SET_Object },
   END_FIELD
};

static ERROR SET_Activate(struct Set *, APTR);
static ERROR SET_ClosingTag(struct Set *, APTR);
static ERROR SET_Free(struct Set *, APTR);
static ERROR SET_GetUnlistedField(struct Set *, struct acGetUnlistedField *);
static ERROR SET_Init(struct Set *, APTR);
static ERROR SET_SetUnlistedField(struct Set *, struct acSetUnlistedField *);

static struct ActionArray SetActions[] = {
   { AC_Activate,         SET_Activate },
   { AC_ClosingTag,       SET_ClosingTag },
   { AC_Free,             SET_Free },
   { AC_GetUnlistedField, SET_GetUnlistedField },
   { AC_Init,             SET_Init },
   { AC_SetUnlistedField, SET_SetUnlistedField },
   { NULL, NULL }
};

/*****************************************************************************
** User Instructions.
*/

static BYTE SetUsage[] = {
"USAGE: set object=name|id field=value ...\n\
\n\
Use the set command to modify the field of any public object\n\
of your choosing.  This powerful command has no restrictions\n\
and is especially useful for debugging.\n\
\n\
EXAMPLE\n\
-------\n\
To change the default colour for buttons and windows to red:\n\
\n\
  set object=gl_colours colour=#ff6060\n"
};

/*****************************************************************************
** Command: Init()
*/

ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   if (CreateObject(ID_MODULE, NULL, &modStrings, NULL,
      FID_Name|TSTRING,   "strings",
      FID_Version|TFLOAT, MODVERSION_STRINGS,
      TAGEND) IS ERR_Okay) {
      if (GetField(modStrings, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&SetClass, NULL,
      FID_BaseClassID|TLONG, ID_SET,
      FID_Version|TFLOAT,    VER_SET,
      FID_Name|TSTRING,      "Set",
      FID_Category|TLONG,    CCF_COMMAND,
      FID_Usage|TPTR,        SetUsage,
      FID_Actions|TPTR,      SetActions,
      FID_Fields|TPTR,       _SetFields,
      FID_Size|TLONG,        sizeof(struct Set),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (modStrings) { acFree(modStrings); modStrings = NULL; }
   if (SetClass)   { acFree(SetClass);   SetClass = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Activate
Short: Activates a Set object.
###END###

*****************************************************************************/

static ERROR SET_Activate(struct Set *Self, APTR Void)
{
   OBJECTPTR object;
   LONG i, j;
   UBYTE str[300], buffer[300];
   DOUBLE Float;
   STRING FieldValue;

   if (Self->ObjectReference[0]) {
      if (StrCopy(Self->ObjectReference, str, sizeof(str)) < sizeof(str)) {
         if (StrTranslate(str, sizeof(str), NULL) IS ERR_Okay) Self->ObjectID = (OBJECTID)StrToInt(str);
         else Self->ObjectID = (OBJECTID)StrToInt(Self->ObjectReference);
      }
      else return(ObjectError(Self, ERH_Activate, ERR_BufferOverflow));
   }

   if (!Self->ObjectID) return(ObjectError(Self, ERH_Activate, ERR_FieldNotSet));

   if (!Self->Fields[0]) {
      DPrintF("@Activate:","[Set:%d] You have not specified any fields for the Set object.", Self->Head.UniqueID);
      return(ERR_FieldNotSet);
   }

   if (AccessObject(Self->ObjectID, 5000, &object) IS ERR_Okay) {
      for (i=0; Self->Fields[i]; i++) {
         for (FieldValue=Self->Fields[i]; *FieldValue; FieldValue++);
         FieldValue++;
         for (j=0; (FieldValue[j]) AND (FieldValue[j] != '['); j++);

         if (FieldValue[j] IS '[') {
            StrCopy(FieldValue, str, sizeof(str));
            StrTranslate(str, sizeof(str), NULL);
            if ((*str IS '=') AND (StrCalculate(str, &Float, buffer, sizeof(buffer)) IS ERR_Okay)) {
               StrCopy(buffer, str, sizeof(str));
            }
            SetFieldVariable(object, Self->Fields[i], str);
         }
         else if ((*FieldValue IS '=') AND (StrCalculate(FieldValue, &Float, buffer, sizeof(buffer)) IS ERR_Okay)) {
            StrCopy(buffer, str, sizeof(str));
            SetFieldVariable(object, Self->Fields[i], str);
         }
         else SetFieldVariable(object, Self->Fields[i], FieldValue);
      }
      ReleaseObject(object);
   }
   else {
      DPrintF("@Activate:","[Set:%d] Failed to access object #%d.", Self->Head.UniqueID, Self->ObjectID);
      return(ERR_AccessObject);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  ClosingTag
Short: If the object is non-static, this action will cause the Set object to activate and then self-destruct.
###END###

*****************************************************************************/

static ERROR SET_ClosingTag(struct Set *Self, APTR Void)
{
   if (Self->Static IS FALSE) {
      Action(AC_Activate, Self, NULL);
      Action(AC_Free, Self, NULL);
   }
   return(ERR_Okay);
}

/*****************************************************************************
** Set: Free
*/

static ERROR SET_Free(struct Set *Self, APTR Void)
{
   WORD i;
   for (i=0; Self->Fields[i]; i++) {
      FreeMemory(Self->Fields[i]);
      Self->Fields[i] = NULL;
   }
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  GetUnlistedField
Short: This action is supported for retrieving unlisted field values.
###END###

*****************************************************************************/

static ERROR SET_GetUnlistedField(struct Set *Self, struct acGetUnlistedField *Args)
{
   STRING fieldvalue;
   LONG i, j;

   if ((!Args) OR (!Args->Buffer)) return(ObjectError(Self, ERH_GetUnlistedField, ERR_Args));

   Args->Buffer[0] = 0;
   for (i=0; Self->Fields[i]; i++) {
      if (StrCompare(Args->Field, Self->Fields[i], NULL, STR_MATCHLENGTH) IS ERR_Okay) {
         for (fieldvalue=Self->Fields[i]; *fieldvalue; fieldvalue++);
         fieldvalue++;

         for (j=0; (fieldvalue[j]) AND (j < Args->Size-1); j++) Args->Buffer[j] = fieldvalue[j];
         Args->Buffer[j++] = 0;

         if (j >= Args->Size) return(ERR_BufferOverflow);
         else return(ERR_Okay);
      }
   }

   DPrintF("@GetVariable:","[Set] The variable \"%s\" does not exist.", Args->Field);

   return(ERR_FieldNotSet);
}

/*****************************************************************************
** Set: Init
*/

static ERROR SET_Init(struct Set *Self, APTR Void)
{
   if ((!Self->ObjectID) AND (!Self->ObjectReference[0])) {
      DPrintF("@Init:","[Set:%d] The Object field has not been set.", Self->Head.UniqueID);
      return(ERR_FieldNotSet);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  SetUnlistedField
Short: This action is supported for setting foreign object field values.
###END###

*****************************************************************************/

static ERROR SET_SetUnlistedField(struct Set *Self, struct acSetUnlistedField *Args)
{
   WORD i, len, index;
   STRING field;

   if ((!Args) OR (!Args->Field) OR (!Args->Value)) return(ERR_Args);

   for (index=0; Self->Fields[index]; index++) {
      if (StrCompare(Args->Field, Self->Fields[index], NULL, STR_MATCHLENGTH) IS ERR_Okay) break;
   }

   if (index < (sizeof(Self->Fields) / sizeof(STRING)) - 1) {
      if (AllocMemory(StrLength(Args->Field) + StrLength(Args->Value) + 2,
          MEM_STRING|MEM_NOCLEAR, &field, NULL) IS ERR_Okay) {

         if (Args->Field[0] IS '?') i = StrCopy(Args->Field+1, field, COPY_ALL) + 1;
         else i = StrCopy(Args->Field, field, COPY_ALL) + 1;

         StrCopy(Args->Value, field + i, COPY_ALL);

         if (field[i] IS '<') {
            len = StrLength(field+i);
            if (field[i+len-1] IS '>') {
               field[i] = '[';
               field[i+len-1] = ']';
            }
         }

         if (Self->Fields[index]) FreeMemory(Self->Fields[index]);
         Self->Fields[index] = field;

         return(ERR_Okay);
      }
      else return(ObjectError(Self, ERH_SetUnlistedField, ERR_AllocMemory));
   }
   else return(ObjectError(Self, ERH_SetUnlistedField, ERR_ArrayFull));
}

/*****************************************************************************

###FIELD###
Name:   Object
Short:  Refers to the object that will receive the field settings.
Type:   STRING/OBJECTID
Status: Read/Write

###DESCRIPTION###
<p>You will need to set this field so that it refers to the object that will
receive the Set object's field settings.  If the Object is not set or is
invalid, the Set object will fail when activated.</p>

<p>If you set this field with a string, the object reference will be
translated to an ID in run-time.</p>

###END###

*****************************************************************************/

static ERROR GET_Object(struct Set *Self, struct Variable *Value)
{
   UBYTE buffer[200];

   if (Value->Type & FD_LARGE) {
      if (Self->ObjectReference[0]) {
         if (StrCopy(Self->ObjectReference, buffer, sizeof(buffer)) < sizeof(buffer)) {
            if (StrTranslate(buffer, sizeof(buffer), NULL) IS ERR_Okay) {
               Value->Large = StrToInt(buffer);
            }
            else Value->Large = StrToInt(Self->ObjectReference);
         }
         else return(ObjectError(Self, ERH_SetField, ERR_BufferOverflow));
      }
      else Value->Large = Self->ObjectID;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
}

static ERROR SET_Object(struct Set *Self, struct Variable *Value)
{
   LONG i;

   if (Value->Type & FD_LARGE) {
      Self->ObjectID = (OBJECTID)Value->Large;
      Self->ObjectReference[0] = 0;
      return(ERR_Okay);
   }
   else if (Value->Type & FD_STRING) {
      if ((!Value->Pointer) OR (!((STRING)Value->Pointer)[0])) return(ERR_Okay);

      for (i=0; (((STRING)Value->Pointer)[i]) AND (((STRING)Value->Pointer)[i] != '['); i++);

      if (((STRING)Value->Pointer)[i] IS '[') {
         if (StrCopy(Value->Pointer, Self->ObjectReference, sizeof(Self->ObjectReference)) < sizeof(Self->ObjectReference)) {
            return(ERR_Okay);
         }
         else return(ObjectError(Self, ERH_SetField, ERR_BufferOverflow));
      }
      else {
         FastFindObject(Value->Pointer, NULL, &Self->ObjectID, 1, NULL);
         return(ERR_Okay);
      }
   }
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));
}

/****************************************************************************

###FIELD###
Name:   Static
Short:  Set to TRUE to make the object static.
Type:   BOOLEAN
Status: Read/Init

###DESCRIPTION###
<p>To make the object static, set this field to TRUE.  If the object is not
made static then it will self-destruct when a closing tag is received.</p>

###END###

****************************************************************************/
